@@ -1,6 +1,8 @@
 #Calculates median scaled difference for observations x given sd's s
 #if s is a scalar function, it is applied to x to obtain an estimate of s
 
+#Index-based msd
+#Still O(n^2) for distance calculation, but a lot faster and lighter on RAM
 msd<-function(x, s=mad , ...) {
         ss <- if(is.function(s)) {
                 rep(s(x, ...), length(x))
@@ -12,13 +14,35 @@ msd<-function(x, s=mad , ...) {
                 }
         }
         
-        m<-abs(outer(x,x,"-")) / outer(ss,ss,FUN=function(a,b) sqrt(a^2+b^2))
-        diag(m) <- NA   #removes deviation from self
+        ss <- ss^2
         
-        return(apply(m, 1, median, na.rm=TRUE))
+        N <- 1:length(x)
         
+        structure( 
+        	sapply(N, function(n) median( abs(x[n] - x[-n])/sqrt(ss[n]+ss[-n]) ) ),
+        	names=names(x)
+        )
 }
 
+#Original msd code retained, commented, for comparison
+#msd<-function(x, s=mad , ...) {
+#        ss <- if(is.function(s)) {
+#                rep(s(x, ...), length(x))
+#        } else {
+#                if(length(s) == 1) {
+#                        rep(s, length(x))
+#                } else {
+#                        s
+#                }
+#        }
+#        
+#        m<-abs(outer(x,x,"-")) / outer(ss,ss,FUN=function(a,b) sqrt(a^2+b^2))
+#        diag(m) <- NA   #removes deviation from self
+#        
+#        return(apply(m, 1, median, na.rm=TRUE))
+#        
+#}
+
 .pmsd.xnorm<-function(q, x, n, sd=1, scale=FALSE) {
         #P for the median of abs(x-X) (optionally /sqrt(2))
         #based on Mood, Graybill and Boes (1974) pp252ff
@@ -39,7 +63,10 @@ msd<-function(x, s=mad , ...) {
         
         ph<-pxnorm(q,x,sd)
         
-        for(j in n.med:(n-1)) Fy <- Fy + choose(n-1,j) * (ph^j) * (1-ph)^(n-j-1)
+        #for(j in n.med:(n-1)) Fy <- Fy + choose(n-1,j) * (ph^j) * (1-ph)^(n-j-1)
+        for(j in n.med:(n-1)) Fy <- Fy + dbeta(ph, j+1, n-j)/n
+			#Added 2016-08-19
+			#beta formulation is considerably more stable at high N
         
         return(Fy)
         
@@ -71,10 +98,11 @@ pmsd<-function(q, n, sd=1, scale=TRUE) {
         
         p <- rep(0, length(q))
         
-        for(i in 1:length(q)) p[i]<-integrate(f, lower=-Inf, upper=Inf, rel.tol = .Machine$double.eps^0.75, 
+        for(i in 1:length(q)) p[i]<-2*integrate(f, lower=0.0, upper=Inf, rel.tol = .Machine$double.eps^0.75, 
                 q=q[i], n=n[i], sd=sd[i], scale=FALSE)$value
                              #Note odd tolerance; pmsd and qmsd are quite inaccurate 
                              #(and qmsd even unstable) with default integrate() tolerance
+                             #Also note (new, 2016) restriction to (0, inf); halves exec time
         return(p)
         
 }
@@ -84,8 +112,10 @@ qmsd<-function(p, n, sd=1, scale=TRUE) {
         #on pmsd.
         #p or n may be vectors
         #scale allows suppression of the scale parameter
+
+        if(any(p >1.0) || any(p < 0.0)) stop("p must be in [0,1]")
         
-        L <- max(length(q), length(n), length(sd))
+        L <- max(length(p), length(n), length(sd))
         p<-rep(p, length.out=L)
         n<-rep(n, length.out=L)
         sd<-rep(sd, length.out=L)
@@ -95,17 +125,18 @@ qmsd<-function(p, n, sd=1, scale=TRUE) {
         q[p==0] <- 0
         q[p==1] <- +Inf
 
-        froot<-function(q, p, n, sd=1, scale) pmsd(q, n, sd, scale)-p
+        froot<-function(qq1, p, n, sd=1, scale) pmsd(qq1/(1-qq1), n, sd, scale)-p
 
-        q.upper <-ifelse(p> 0.9999, 10,4) #saves a step
+        qq1.upper <-ifelse(p > 0.9999, 1.0, 0.8) #saves a step
         
         for(i in 1:length(p)) {
                 
                 if(q[i]<0) { #Note explicit values above
                              #Also note odd tolerance; qmsd is quite inaccurate with default tolerance
-                        q[i] <- uniroot(froot, interval=c(0,q.upper[i]), tol=.Machine$double.eps^0.75,
+                        qq1 <- uniroot(froot, interval=c(0,qq1.upper[i]), tol=.Machine$double.eps^0.75,
                                 p=p[i], n=n[i], 
                                 sd=sd[i], scale=scale)$root
+                        q[i] <- qq1/(1-qq1)
                 }
         }
 
